#include "image.h"

#include <png.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <fstream>
#include <iostream>
#include <iomanip>

bool load(const char *file, Image &image)
{
	std::string name = file;
	if (name.size() < 4)
		return false;
	name = name.substr(name.size() - 4);
	
	if (name == ".png" || name == ".PNG")
		return loadPNG(file, image);
	else if (name == ".tga" || name == ".TGA")
		return loadTGA(file, image);
	else
	{
		std::cout << "Unknown image type (\"" << file << "\")." << std::endl;
		return false;
	}
}

bool loadTGA(const char *fileName, Image &image)
/* This code originates from the EFnet #OpenGL FAQ, available here:
	http://www.geocities.com/SiliconValley/Park/5625/opengl/ (8/14/2002) */
{
	// Targa data type -- uncompressed, RGBA image; element
	// 3 is always 2.
	unsigned char  szTargaMagic[12] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
	unsigned char  szFileMagic[12];

	// Header data.
	unsigned char szHeader[6];
	unsigned int  uBytesPerPixel;
	unsigned int  uImageSize;

	std::ifstream	file(fileName, std::ios::binary|std::ios::in);

	// Start reading in the header, make sure we
	// have the right Targa data type.

	if (!file.read((char *)&szFileMagic, sizeof(szFileMagic)))
		return false;
	if (memcmp(szTargaMagic, szFileMagic, sizeof(szTargaMagic)))
		return false;
	if (!file.read((char *)szHeader, sizeof(szHeader)))
		return false;

 	// Determine width and height from header data.

	image.width  = szHeader[1] * 256 + szHeader[0];
	image.height = szHeader[3] * 256 + szHeader[2];
	 
	 // Determine color bit depth of the image (24 or 32).
	image.bpp = szHeader[4];
	uBytesPerPixel = image.bpp / 8;
	uImageSize = image.width * image.height * uBytesPerPixel;

  if(image.width <= 0 || image.height <= 0 || (image.bpp != 24 && image.bpp != 32))
    return false;

  image.data.resize(uImageSize);

  if (image.data.size() != uImageSize)
    return false;

  {
    file.read(reinterpret_cast<char *>(&*image.data.begin()), uImageSize);
    unsigned int i;

    for(i = 0; i < uImageSize; i += uBytesPerPixel)
    {
      unsigned int uTmp;
      uTmp = image.data[i];
      image.data[i] = image.data[i + 2];
      image.data[i + 2] = uTmp;
    }
  }

  return true;
}

bool loadPNG(const char *file, Image &image)
{
	image.width = image.height = image.bpp = 0;
	image.data.resize(0);

	FILE *fp = fopen(file, "rb");
	if (!fp)
		return false;
	unsigned char header[8];
	fread(header, 1, 8, fp);
	if (png_sig_cmp(header, 0, 8))
		return false; // Not a PNG file.

	png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if (!png_ptr)
		return false;

	png_infop info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	{
		png_destroy_read_struct(&png_ptr, 0, 0);
		return false;
	}

	png_infop end_info = png_create_info_struct(png_ptr);
	if (!end_info)
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)0);
		return false;
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
		fclose(fp);
		return false;
	}

	png_init_io(png_ptr, fp);
	png_set_sig_bytes(png_ptr, 8);
	png_read_info(png_ptr, info_ptr);
	int bit_depth;
	int color_type;
	png_get_IHDR(png_ptr, info_ptr, (png_uint_32 *)&image.width, (png_uint_32 *)&image.height, &bit_depth, &color_type, 0, 0, 0);

	if (bit_depth == 16)
		png_set_strip_16(png_ptr);
	if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
		png_set_expand_gray_1_2_4_to_8(png_ptr);

	if (color_type == PNG_COLOR_TYPE_GRAY)
		image.bpp = 8;
	else if (color_type == PNG_COLOR_TYPE_RGB)
		image.bpp = 24;
	else if (color_type == PNG_COLOR_TYPE_RGBA)
		image.bpp = 32;
	else
	{	// We don't care enough about other formats to load them.
		png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
		fclose(fp);
		return false;
	}

	image.data.resize(image.width * image.height * image.bpp / 8);

	std::vector<png_bytep> row_pointers(image.height);

	for (int row = 0; row < image.height; row++)
		row_pointers[image.height - row - 1] = (png_byte *)(&*image.data.begin() + row * image.width * (image.bpp / 8));

	png_read_image(png_ptr, &*row_pointers.begin());

	png_read_end(png_ptr, end_info);

	png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);

	return true;
}
