#include "physics.h"

#include <cmath>
#include <iostream>

#include "object.h"

void Physics::determineFinalState(Environment env)
{
	// This needs to be updated to use the same calculations as advanceState

	while (!allObjectsStationary(*env.objects)) // I'm not sure this will ever be useful, but we will need a way to end this loop.
		advanceState(env);
}

void Physics::advanceState(Environment env, float timeToAdvance)
{
	static float moreTime = 0.0f;

	for (moreTime += timeToAdvance; moreTime > timeStep; moreTime -= timeStep)
		advanceState(env);
}

void Physics::advanceState(Environment env)
{
	// This loop is executed approximately (yet consistently) every timeStep seconds.

	std::list<Object *> &objects = *env.objects;
	const Wall &wall = *env.wall;
	Terrain &terrain = *env.terrain;
	
	vertex bounceFactor;
	Object *a; // This is rather ugly.  a and b are for sphere collisions, c is for wall collisions, and d is for terrain...
	Object *b;
	Object *c;
	Object *d;

	a = b = c = d = 0;

	// Erase any objects marked as deleted, first.
	std::list<Object *>::iterator del = env.objects->begin();
	while (del != env.objects->end())
	{
		if ((*del)->deleted)
		{
			Object *p = *del;
			del = env.objects->erase(del);
			delete p;
		}
		else
			del++;
	}
	
	// Here we find all possible collisions...we'll only be concerned with the first, but it might be one of these three.
	float t1 = getNextCollision(objects, a, b);
	float t2 = getNextCollision(objects, wall, c, bounceFactor);
	float t3 = getNextCollision(objects, terrain, d);

	if (t1 >= 0 && t1 <= t2 && t1 <= t3 && a && b)
	{
		// Two objects (spheres) hitting each other.
		advanceObjects(env, t1);
		if (t1 <= timeStep && a && b)
		{
			reactToCollision(*a, *b);
			a->contactEvent(env);
			b->contactEvent(env);
			std::cout << "bounce sphere " << a->id() << " sphere " << ", " << b->id() << std::endl;
			std::cout << "resting:  " << a->resting << ", " << b->resting << std::endl;
			std::cout << "fixed:    " << a->fixed << ", " << b->fixed << std::endl;
			std::cout << "immobile: " << a->immobile << ", " << b->immobile << std::endl;
		}
	}
	else if (t2 >= 0 && t2 <= t3)
	{
		// An object hitting a wall.
		advanceObjects(env, t2);
		if (t2 < timeStep && c)
		{
			std::cout << "bounce sphere wall" << c->id() << std::endl;
			c->velocity.x *= bounceFactor.x;
			c->velocity.y *= bounceFactor.y;
			c->velocity.z *= bounceFactor.z;
		}
		// There should be all kinds of crazy walls
	}
	else if (t3 >= 0 /*&& t3 <= t1 && t3 <= t2*/)
	{
		// And object hitting the ground.
		advanceObjects(env, t3);
		if (t3 < timeStep && d)
		{
			std::cout << "bounce sphere terrain " << d->id() << std::endl;
			d->velocity.x = d->velocity.y = d->velocity.z = 0.0f;
			d->resting = true; // There could be some bounceage, but we don't need it.
			
			d->contactEvent(env);
		}
	}
}

void Physics::advanceObjects(Environment env, float timeToAdvance)
{
	// Move all of the objects and update the velocities based on acceleration.
	for (std::list<Object *>::iterator i = env.objects->begin(); i != env.objects->end(); i++)
	{
		if ((*i)->fixed)
			(*i)->resting = true;
			
		/* I'm not sure that this belongs just here, but if we find an object that "wants" to move and it's marked as resting, then we should
			take away its resting status so that it can move again. */
		if ((*i)->active && (*i)->resting && !(*i)->fixed)
		{
			const float minVelocity = 0.1f;
			if (dotProduct((*i)->velocity, (*i)->velocity) >= minVelocity * minVelocity)
				(*i)->resting = false;
		}
		
		if ((*i)->active && !(*i)->resting && !(*i)->fixed)
		{
			(*i)->position += timeToAdvance * (*i)->velocity;
			(*i)->velocity += timeToAdvance * gravity;
		}
	}

	// Send all of the objects an "advance" event.
	for (std::list<Object *>::reverse_iterator j = env.objects->rbegin(); j != env.objects->rend(); j++)
		if ((*j)->active)
			(*j)->advanceEvent(env, timeToAdvance);
}

// Returns a value in the range [0, timeStep) indicating the time of the next collision or timeStep if no collisions are found.
float Physics::getNextCollision(std::list<Object *> &objects, const Wall &wall, Object *&a, vertex &bounceFactor)
{
	vertex one = {1, 1, 1};
	float uMin = timeStep;
	bounceFactor = one;

	for (std::list<Object *>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		if ((*i)->active)
		{
			if ((*i)->velocity.x)
			{
				float t = timeStep;
				if ((*i)->velocity.x < 0)
					t = (wall.left - (*i)->position.x) / (*i)->velocity.x;
				if ((*i)->velocity.x > 0)
					t = (wall.right - (*i)->position.x) / (*i)->velocity.x;
				if (t < uMin)
				{
					uMin = t;
					a = *i;
					bounceFactor.x = -1.0f;
					bounceFactor.y = bounceFactor.z = 1.0f;
				}
			}
			if ((*i)->velocity.y)
			{
				float t = timeStep;
				if ((*i)->velocity.y < 0)
					t = (wall.bottom - (*i)->position.y) / (*i)->velocity.y;
				if ((*i)->velocity.y > 0)
					t = (wall.top - (*i)->position.y) / (*i)->velocity.y;
				if (t < uMin)
				{
					uMin = t;
					a = *i;
					bounceFactor.y = -1.0f;
					bounceFactor.x = bounceFactor.z = 1.0f;
				}
			}
			if ((*i)->velocity.z)
			{
				float t = timeStep;
				if ((*i)->velocity.z < 0)
					t = (wall.back - (*i)->position.z) / (*i)->velocity.z;
				if ((*i)->velocity.z > 0)
					t = (wall.front - (*i)->position.z) / (*i)->velocity.z;
				if (t < uMin)
				{
					uMin = t;
					a = *i;
					bounceFactor.z = -1.0f;
					bounceFactor.x = bounceFactor.y = 1.0f;
				}
			}
		}
	}

	if (uMin < 0)
		uMin = 0.0f;
	return uMin;
}

// Returns a value in the range [0, timeStep) indicating the time of the next collision or timeStep if no collisions are found.
float Physics::getNextCollision(std::list<Object *> &objects, Object *&a, Object *&b)
{
	std::list<Object *>::iterator i;
	std::list<Object *>::iterator j;
	float uMin = timeStep;

	a = b = 0;

	for (i = objects.begin(); i != objects.end(); i++)
	{
		for (j = i, j++; j != objects.end(); j++)
		{
			if ((*i)->active && (*j)->active && !((*i)->resting && (*j)->resting) && !((*i)->resting && (*j)->immobile) && !((*j)->resting && (*i)->immobile))
			{
				float u = sphereSphereCollision(**i, **j, timeStep);
				if (u < uMin)
				{
					uMin = u;
					a = *i;
					b = *j;
				}
			}
		}
	}

	return uMin;
}

// Returns a value in the range [0, timeStep) indicating the time of the next collision or timeStep if no collisions are found.
float Physics::getNextCollision(std::list<Object *> &objects, Terrain &terrain, Object *&a)
{
	/* Please note that this function is wrong.  It is very very wrong.  I most certainly
		didn't even try to make it right.  It's wrong.  The function, that is...it's wrong.
		It kind of looks like it works at a brief glance, but then you look closely and it's
		really very wrong.  That's just the thing.  This function is wrong. Please fix it. */
	vertex scale = terrain.scale();
	float uMin = timeStep;

	for (std::list<Object *>::iterator i = objects.begin(); i != objects.end(); i++)
	{
		if ((*i)->active && !(*i)->resting)
		{
			vertex p = (*i)->position + timeStep * (*i)->velocity;
			int x = static_cast<int>(p.x / scale.x * terrain.width());
			int y = static_cast<int>(p.z / scale.z * terrain.height());
			if (x < 0) x = 0;
			if (y < 0) y = 0;
			if (static_cast<unsigned int>(x) >= terrain.width()) x = terrain.width() - 1;
			if (static_cast<unsigned int>(y) >= terrain.height()) y = terrain.height() - 1;
			if (((p.y - (*i)->radius) < terrain.maxHeight(x, y)))
			{
				float u = 0.0f;
	
				if ((*i)->velocity.y)
					u = (terrain.maxHeight(x, y) - (*i)->position.y + (*i)->radius) / (*i)->velocity.y;
				if (u < uMin)
				{
					a = *i;
					uMin = u;
					if (uMin < 0) uMin = 0;
				}
			}
		}
	}

	return uMin;
}

void Physics::reactToCollision(Object &a, Object &b)
{
	// We can assume that a and b are touching and that they collided on the radial between them.

	if (a.fixed || b.fixed)
	{
		/*Object &fixed = a.fixed?a:b;
		Object &moving = a.fixed?b:a;

		
		vertex c = moving.position - fixed.position;
		float absc = abs(c);
		if (absc == 0) absc = 1;
		vertex nc = c / absc;

		float vos = (moving.velocity.x || moving.velocity.y || moving.velocity.z)?(dotProduct(moving.velocity, c) / absc):0;
		vertex vor = moving.velocity - 2 * vos * nc;
		
		moving.velocity = moving.velocity - 2 * vor;

		std::cout << "reacting to collision..." << moving.id() << std::endl;*/

		return;
	}

	vertex vao = a.velocity;
	vertex vbo = b.velocity;

	float ma = a.mass;
	float mb = b.mass;
	float elasticity = a.elasticity * b.elasticity;

	vertex c = a.position - b.position;

	float absc = abs(c);
	if (absc == 0) absc = 1.0f;

	vertex nc = c / absc;

	float vaos = (vao.x || vao.y || vao.z)?(dotProduct(vao, c) / absc):0;
	vertex vaor = vao - vaos * nc;
	float vbos = (vbo.x || vbo.y || vbo.z)?(dotProduct(vbo, c) / absc):0;
	vertex vbor = vbo - vbos * nc;

	a.velocity = vaor + elasticity * nc * (2 * mb * vbos + vaos * (ma - mb)) / (ma + mb);
	b.velocity = vbor + elasticity * nc * (2 * ma * vaos + vbos * (mb - ma)) / (ma + mb);
}

bool Physics::allObjectsStationary(const std::list<Object *> &objects)
{
	std::list<Object *>::const_iterator i;

	for (i = objects.begin(); i != objects.end(); i++)
		if (!(*i)->resting)
			return false;

	return true;
}

float Physics::sphereSphereCollision(const Object &a, const Object &b, float t)
{
	float rv = relativeVelocity(a.position, b.position, a.velocity, b.velocity);

	if (dotProduct(a.position - b.position, a.position - b.position) <= (a.radius + b.radius) * (a.radius + b.radius) && rv < 0.0f)
	{
		return 0;
	}
	if (rv < 0)
	{
		float u = sphereSphereIntersection(a, b, t) * t;
		return u;
	}
	else
		return t;
}

/* Returns a value from 0 to 1 */
float Physics::sphereSphereIntersection(const Object &p, const Object &q, float t)
{
	float radius = p.radius + q.radius;
	vertex A1 = p.position;
	vertex B1 = q.position;
	vertex A2 = p.position + p.velocity * t;
	vertex B2 = q.position + q.velocity * t;
	vertex A = (A2 - A1) - (B2 - B1);
	float a = A.x * A.x + A.y * A.y + A.z * A.z;
	float b = dotProduct(2 * (A1 - B1), ((A2 - A1) - (B2 - B1)));
	vertex C = (A1 - B1);
	float c = C.x * C.x + C.y * C.y + C.z * C.z - radius * radius;
	float descr = b * b - 4 * a * c;

	if (descr < 0)
	{ // Imaginary roots
		return 1;
	}
	else if (descr == 0)
	{ // Double root
		float root = -b / (2 * a);
		if (root >= 0 && root < 1)
			return root;
		else
			return 1;
	}
	else
	{ // Two unique roots
		float sqrtdescr = static_cast<float>(sqrt(descr));
		float root1 = (-b - sqrtdescr) / (2 * a);
		float root2 = (-b + sqrtdescr) / (2 * a);

		//  We know root1 < root2
		if (root1 >= 0 && root1 < 1)
			return root1;
		if (root2 >= 0 && root2 < 1)
			return root2;
		return 1;
	}
}
