#include "skeleton.h"

#include "xmlparser.h"

#include <stdlib.h>

#include <stack>
#include <vector>
#include <string>

class SkeletonParser : public XMLParser
{
	SkeletonNode &skeleton;
	std::stack<SkeletonNode *> nodes;
	std::vector<std::string> tagStack;
public:
	SkeletonParser(SkeletonNode &newSkeleton) :skeleton(newSkeleton) {}
	void start(void *userData, const XML_Char *name, const XML_Char **atts);
	void end(void *userData, const XML_Char *name);
	void data(void *userData, const XML_Char *s, int len);
};

void SkeletonParser::start(void *userData, const XML_Char *name, const XML_Char **atts)
{
	tagStack.push_back(name);

	if (!nodes.size())
	{
		if (std::string(name) == "skeleton")
		{
			nodes.push(&skeleton);
			skeleton.translation.x = skeleton.translation.y = skeleton.translation.z = 0.0f;
			skeleton.rotation.x = skeleton.rotation.y = skeleton.rotation.z = 0.0f;
			skeleton.scale.x = skeleton.scale.y = skeleton.scale.z = 1.0f;
		}
	}
	else if (std::string(name) == "child")
	{
		SkeletonNode &current = *nodes.top();
		current.child.resize(current.child.size() + 1);
		nodes.push(&current.child[current.child.size() - 1]);
		SkeletonNode &n = current.child[current.child.size() - 1];
		n.translation.x = n.translation.y = n.translation.z = 0.0f;
		n.rotation.x = n.rotation.y = n.rotation.z = 0.0f;
		n.scale.x = n.scale.y = n.scale.z = 1.0f;
	}
}

void SkeletonParser::end(void *userData, const XML_Char *name)
{
	tagStack.pop_back();

	if (std::string(name) == "child")
		nodes.pop();
}

void SkeletonParser::data(void *userData, const XML_Char *s, int len)
{
	std::string data(s, s + len);
	std::string tag;
	if (tagStack.size()) tag = *tagStack.rbegin();
	if (nodes.size())
	{
		if (tag == "id")
			nodes.top()->id = data;
		else if (tag == "mesh")
			nodes.top()->mesh = data;
		else if ((tag == "x" || tag == "y" || tag == "z") && tagStack.size() > 1)
		{
			std::string prevTag = tagStack[tagStack.size() - 2];
			vertex *vector = 0;

			if (prevTag == "translate")
				vector = &nodes.top()->translation;
			else if (prevTag == "rotate")
				vector = &nodes.top()->rotation;
			else if (prevTag == "scale")
				vector = &nodes.top()->scale;

			if (tag == "x")
				vector->x = static_cast<float>(atof(s));
			else if (tag == "y")
				vector->y = static_cast<float>(atof(s));
			else if (tag == "z")
				vector->z = static_cast<float>(atof(s));
		}
	}
}

bool load(const char *filename, SkeletonNode &skeleton)
{
	SkeletonParser parser(skeleton);
	return parser.parse(filename);
}

tMatrix transformationMatrix(const SkeletonNode &skeletonNode)
{
	tMatrix T;
	
	T.translate(skeletonNode.translation);
	T.rotate(skeletonNode.rotation);
	T.scale(skeletonNode.scale);

	return T;
}

const SkeletonNode *findChildPath(const SkeletonNode &skeleton, const char *joint)
{
	for (std::vector<SkeletonNode>::const_iterator i = skeleton.child.begin(); i != skeleton.child.end(); i++)
	{
		if (i->id == joint)
			return &*i;
		if (findChildPath(*i, joint))
			return &*i;
	}
	
	return 0;
}

bool getTransformationMatrix(const SkeletonNode &skeleton, const char *joint, tMatrix &t)
{
	t = t * transformationMatrix(skeleton);
	
	if (skeleton.id == joint)
		return true;
	
	const SkeletonNode *nextNode = &skeleton;

	while ((nextNode = findChildPath(*nextNode, joint)))
	{
		t = t * transformationMatrix(*nextNode);
		if (nextNode->id == joint)
			return true;
	}

	return false;
}

/* This function is intended to be used to find the tip of a barrel. */
bool transformVertex(const SkeletonNode &skeleton, vertex &vert, const char *joint)
{
	tMatrix T;
	if (getTransformationMatrix(skeleton, joint, T))
	{
		vert = T * vert;
		return true;
	}

	return false;
}
