#include "terrain.h"

#include "random.h"

#include <math.h>

Terrain::Terrain(unsigned int width, unsigned int height, float xScale, float yScale, float zScale)
: Matrix<TerrainValue>(width, height)
{
	terrainScale.x = xScale;
	terrainScale.y = yScale;
	terrainScale.z = zScale;
	clear(0);
}

void Terrain::clear(float zero)
{
	unsigned int terrainWidth = width();
	unsigned int terrainHeight = height();
	unsigned int x;
	unsigned int y;
	vertex normal = {0, 1, 0};

	for (x = 0; x < terrainWidth; x++)
		for (y = 0; y < terrainHeight; y++)
		{
			at(x, y).height = zero;
			at(x, y).normal = normal;
		}
}

void Terrain::generate(int iterations, float floor, float ceiling)
{
	Random random;
	unsigned int terrainWidth = width();
	unsigned int terrainHeight = height();
	
	if (!terrainWidth || !terrainHeight) return;

	
	while (iterations--)
	{
		int s = static_cast<int>((terrainWidth > terrainHeight) ? terrainHeight : terrainWidth);
		s /= 5;
		int x = random.number(0, static_cast<int>(terrainWidth));
		int y = random.number(0, static_cast<int>(terrainHeight));
		int r = random.number(2, s);
		float h = float(random.number(0, 101)) / 100.0f;

		int p;
		int q;
		int i;
		int j;
		for (p = -r; p < r; p++)
		{
			for (q = -r; q < r; q++)
			{
				i = x + p;
				j = y + q;
				if (i >= 0 && j >= 0 && i < int(terrainWidth) && j < int(terrainWidth) && (p * p + q * q <= r * r))
				{
					at(i, j).height += h * (float(r * r) - float(p * p) - float(q * q));
				}
			}
		}
	}

	float max = at(0, 0).height;
	float min = at(0, 0).height;

	for (unsigned int x = 0; x < terrainWidth; x++)
		for (unsigned int y = 0; y < terrainHeight; y++)
		{
			at(x, y).height *= at(x, y).height;
			if (at(x, y).height > max)
				max = at(x, y).height;
			if (at(x, y).height < min)
				min = at(x, y).height;
		}

	for (unsigned int x = 0; x < terrainWidth; x++)
		for (unsigned int y = 0; y < terrainHeight; y++)
			at(x, y).height = (at(x, y).height - min) / (max - min) * (ceiling - floor) + floor;

	generateNormals(0, 0, terrainWidth - 1, terrainHeight - 1);
	generateTexture(0, 0, terrainWidth - 1, terrainHeight - 1);
}

void Terrain::generateNormals(unsigned int left, unsigned int bottom, unsigned int right, unsigned int top)
{
	Matrix<vertex> normalOne(right - left + 1, top - bottom + 1);
	Matrix<vertex> normalTwo(right - left + 1, top - bottom + 1);
	unsigned int x;
	unsigned int y;

	for (x = 0; x < (right - left); x++)
	{
		for (y = 0; y < (top - bottom); y++)
		{
			normalOne.at(x, y) = surfaceNormal(x + left, y + bottom, true);
			normalTwo.at(x, y) = surfaceNormal(x + left, y + bottom, false);
		}
	}

	for (x = 0; x <= right - left; x++)
	{
		for (y = 0; y <= top - bottom; y++)
		{
			vertex sum = {0, 0, 0};
			int count = 0;
			if (x >= 0 && y >= 0 && x < width() && y < height())
			{
				sum += normalOne.at(x, y) + normalTwo.at(x, y);
				count += 2;
			}
			if (x - 1 >= 0 && y - 1 >= 0 && x - 1 < width() && y - 1 < height())
			{
				 sum += normalOne.at(x - 1, y - 1) + normalTwo.at(x - 1, y - 1);
				 count += 2;
			}
			if (x - 1 >= 0 && y >= 0 && x - 1 < width() && y < height())
			{
				sum += normalTwo.at(x - 1, y);
				count++;
			}
			if (x >= 0 && y - 1 >= 0 && x < width() && y - 1 < height())
			{
				sum += normalOne.at(x, y - 1);
				count++;
			}
			if (count)
				sum /= static_cast<float>(count);
			else
				std::cout << "No count" << std::endl;
			at(x + left, y + bottom).normal = sum;
		}
	}
}

vertex Terrain::surfaceNormal(unsigned int x, unsigned int y, bool one)
{
	if (one)
	{
		vertex a = {terrainScale.x / width(), at(x + 1, y + 1).height - at(x, y + 1).height, 0.0f};
		vertex b = {0.0f, at(x, y).height - at(x, y + 1).height, -terrainScale.z / height()};
		return normalize(crossProduct(a, b));
	}
	else
	{
		vertex c = {-terrainScale.x / width(), terrainScale.x * (at(x, y).height - at(x + 1, y).height), 0.0f};
		vertex d = {0.0f, terrainScale.y * (at(x + 1, y + 1).height - at(x + 1, y).height), terrainScale.z / height()};
		return normalize(crossProduct(c, d));
	}
}

void Terrain::scale(float x, float y, float z)
{
	terrainScale.x = x;
	terrainScale.y = y;
	terrainScale.z = z;
	generateNormals(0, 0, width() - 1, height() - 1);
	generateNormals(0, 0, width() - 1, height() - 1);
}

vertex Terrain::scale()
{
	return terrainScale;
}

float Terrain::maxHeight(unsigned int x, unsigned int y)
{
	float m = at(x, y).height;

	for (unsigned int i = x; i < x + 1; i++)
	{
		for (unsigned int j = y; j < y + 1; j++)
		{
			if (x >= 0 && x < width() && y >= 0 && y < height() && at(i, j).height > m)
			{
				m = at(i, j).height;
			}
		}
	}
	return m * terrainScale.y;
}

float Terrain::minHeight(unsigned int x, unsigned int y)
{
	float m = at(x, y).height;

	for (unsigned int i = x; i < x + 1; i++)
	{
		for (unsigned int j = y; j < y + 1; j++)
		{
			if (x >= 0 && x < width() && y >= 0 && y < height() && at(i, j).height < m)
			{
				m = at(i, j).height;
			}
		}
	}
	return m * terrainScale.y;
}

void Terrain::sphereDeform(const vertex &position, float radius)
{
	int sx1 = static_cast<int>((position.x - radius) / terrainScale.x * width() - 1.0f);
	int sx2 = static_cast<int>((position.x + radius) / terrainScale.x * width() + 1.0f);
	int sy1 = static_cast<int>((position.z - radius) / terrainScale.z * height() - 1.0f);
	int sy2 = static_cast<int>((position.z + radius) / terrainScale.z * height() + 1.0f);

	if (sx1 < 0) sx1 = 0; if (sx1 >= static_cast<int>(width())) sx1 = static_cast<int>(width() - 1);
	if (sx2 < 0) sx2 = 0; if (sx2 >= static_cast<int>(width())) sx2 = static_cast<int>(width() - 1);
	if (sy1 < 0) sy1 = 0; if (sy1 >= static_cast<int>(height())) sy1 = static_cast<int>(height() - 1);
	if (sy2 < 0) sy2 = 0; if (sy2 >= static_cast<int>(height())) sy2 = static_cast<int>(height() - 1);

	unsigned int x1 = static_cast<unsigned int>(sx1);
	unsigned int x2 = static_cast<unsigned int>(sx2);
	unsigned int y1 = static_cast<unsigned int>(sy1);
	unsigned int y2 = static_cast<unsigned int>(sy2);

	unsigned int x;
	unsigned int y;

	for (x = x1; x <= x2; x++)
	{
		for (y = y1; y <= y2; y++)
		{
			float fx = -position.x + x * terrainScale.x / width();
			float fy = -position.z + y * terrainScale.z / height();
			float h = at(x, y).height * terrainScale.y;
			float f = radius * radius - fx * fx - fy * fy;
			if (f > 0)
			{
				float dh = static_cast<float>(sqrt(f));

				if (h > position.y + dh) at(x, y).height += (h - (position.y + dh)) / terrainScale.y;
				if (h > position.y - dh) at(x, y).height -= (h - (position.y - dh)) / terrainScale.y;
				if (at(x, y).height < 0) at(x, y).height = 0.0f;
			}
		}
	}

	generateNormals(0, 0, width() - 1, height() - 1); //x1, y1, x2, y2);
	generateTexture(0, 0, width() - 1, height() - 1);
}

void Terrain::generateTexture(unsigned int left, unsigned int bottom, unsigned int right, unsigned int top)
{
	/* The purpose of this function is to put different textures on different parts of the terrain.
		Mountan peaks become snowy, lower areas are grassy...stuff like that. */

	Matrix<int> heights(width(), height());
	std::list<TextureInfo>::iterator level;
	std::list<TextureInfo>::iterator previous;
	unsigned int x;
	unsigned int y;
	int height = 1;

	for (x = left; x < right; x++)
		for (y = bottom; y < top; y++)
			at(x, y).texture = 0; // Initialize the thing to zero, at first.

	for (level = textureInfo.begin(), previous = textureInfo.end(); level != textureInfo.end(); previous = level++, height++)
	{
		// The first step is to put a nice edge around any tiles from the previous iteration.
		for (x = left; x <= right; x++)
		{
			for (y = bottom; y <= top; y++)
			{
				if (!at(x, y).texture)
				{
					bool corners[4] = {false, false, false, false};
	
					if (x > left)
					{
						if (heights.at(x - 1, y) == height - 1)
							corners[0] = corners[3] = true;
						if (y > bottom && heights.at(x - 1, y - 1) == height - 1)
							corners[0] = true;
						if (y < top && heights.at(x - 1, y + 1) == height - 1)
							corners[3] = true;
					}
					if (y > bottom)
					{
						if (heights.at(x, y - 1) == height - 1)
							corners[0] = corners[1] = true;
					}
					if (x < right)
					{
						if (heights.at(x + 1, y) == height - 1)
							corners[1] = corners[2] = true;
						if (y > bottom && heights.at(x + 1, y - 1) == height - 1)
							corners[1] = true;
						if (y < top && heights.at(x + 1, y + 1) == height - 1)
							corners[2] = true;
					}
					if (y < top)
					{
						if (heights.at(x, y + 1) == height - 1)
							corners[2] = corners[3] = true;
					}
	
					int count = 0;
					int corner;
					for (corner = 0; corner < 4; corner++)
						if (corners[corner])
							count++;
	
					if (count && previous != textureInfo.end()) // previous is set to textureInfo.end() previously on the first iteration
					{
						switch (count)
						{
						case 4:
							at(x, y).texture = previous->textures[0];
							at(x, y).orientation = 0;
							break;
						case 3:
							at(x, y).texture = previous->textures[3];
							if (!corners[1]) at(x, y).orientation = 0;
							if (!corners[2]) at(x, y).orientation = 1;
							if (!corners[3]) at(x, y).orientation = 2;
							if (!corners[0]) at(x, y).orientation = 3;
							break;
						case 2: // There's a case with two opposite corners at a different level that isn't accounted for.
							if (corners[0] == corners[2]) // The diagonal case
							{
								at(x, y).texture = previous->textures[4];
								if (corners[0]) at(x, y).orientation = 1;
								else at(x, y).orientation = 0;
							}
							else
							{
								at(x, y).texture = previous->textures[2];
								if (!corners[3] && corners[0]) at(x, y).orientation = 0; 
								if (!corners[0] && corners[1]) at(x, y).orientation = 1;
								if (!corners[1] && corners[2]) at(x, y).orientation = 2;
								if (!corners[2] && corners[3]) at(x, y).orientation = 3;
							}
							break;
						case 1:
							at(x, y).texture = previous->textures[1];
							if (corners[1]) at(x, y).orientation = 0;
							if (corners[2]) at(x, y).orientation = 1;
							if (corners[3]) at(x, y).orientation = 2;
							if (corners[0]) at(x, y).orientation = 3;
							break;
						}

						heights.at(x, y) = height;
					}
				}
			}
		}

		for (x = left; x < right; x++)
		{
			for (y = bottom; y < top; y++)
			{
				std::list<TextureInfo>::iterator next = level;
				next++;
				
				if (((at(x, y).height + at(x + 1, y).height + at(x, y + 1).height + at(x + 1, y + 1).height) / 4 < level->height || next == textureInfo.end())&& !at(x, y).texture)
				{
					heights.at(x, y) = height;
					at(x, y).texture = level->textures[0];
					at(x, y).orientation = 0;
				}
			}
		}
	}
}
